/*
 * Copyright 2007 Philip Jones, EMBL-European Bioinformatics Institute
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 *
 * For further details of the mydas project, including source code,
 * downloads and documentation, please see:
 *
 * http://code.google.com/p/mydas/
 *
 */

package uk.ac.ebi.mydas.datasource;

import uk.ac.ebi.mydas.exceptions.DataSourceException;
import uk.ac.ebi.mydas.exceptions.BadReferenceObjectException;
import uk.ac.ebi.mydas.exceptions.CoordinateErrorException;
import uk.ac.ebi.mydas.model.DasAnnotatedSegment;

/**
 * Created Using IntelliJ IDEA.
 * Date: 04-May-2007
 * Time: 14:34:22
 *
 * @author Phil Jones, EMBL-EBI, pjones@ebi.ac.uk
 *
 * If you implement this interface, this will allow you to take control
 * of filtering by start / stop coordinates in your AnnotationDataSource.
 *
 * This is useful if your DAS source includes massive segments, otherwise don't bother - just implement a
 * {@link AnnotationDataSource} and the mydas servlet will take care of filtering the features by coordinates.
 */
public interface RangeHandlingAnnotationDataSource extends AnnotationDataSource{

    /**
     * This method returns a List of DasAnnotatedSegment objects, describing the annotated segment and the features
     * of the segmentId passed in as argument.
     * @param segmentId being the reference of the segment requested in the DAS request (not including
     * start and stop coordinates)
     *
     * If your datasource implements only this interface,
     * the MydasServlet will handle restricting the features returned to
     * the start / stop coordinates in the request and you will only need to
     * implement this method to return Features.  If on the other hand, your data source
     * includes massive segments, you may wish to implement the {@link uk.ac.ebi.mydas.datasource.RangeHandlingAnnotationDataSource}
     * interface.  It will then be the responsibility of your AnnotationDataSource plugin to
     * restrict the features returned for the requested range.
     * @param start being the requested start coordinate after which features should be returned.
     * @param stop being the requested stop coordinate before which features should be returned.
     *
     * @return A DasAnnotatedSegment object.  This describes the segment that is annotated, limited
     * to the information required for the /DASGFF/GFF/SEGMENT element.  References a Collection of
     * DasFeature objects.   Note that this is a basic Collection - this gives you complete control over the details
     * of the Collection type - so you can create your own comparators etc.
     * @throws BadReferenceObjectException should be thrown if the segment requested does not
     * exist in this data source.
     * @throws DataSourceException to wrap any other types of exception that may be
     * generated by your data source, such as SQLExceptions or IOExceptions.
     * @throws uk.ac.ebi.mydas.exceptions.CoordinateErrorException should be thrown if the requested coordinates
     * fall outside the bounds of the requested segments.
     */
    public DasAnnotatedSegment getFeatures(String segmentId, int start, int stop) throws BadReferenceObjectException, CoordinateErrorException, DataSourceException;

}
